@@ -1,6 +1,8 @@
 // Copyright (c) 2020-2023 Christopher Davis
-// Copyright (c) 2022-2023 Sophie Herold
+// Copyright (c) 2022-2024 Sophie Herold
 // Copyright (c) 2022-2023 Sabri Ünal
+// Copyright (c) 2023 Matteo Nardi
+// Copyright (c) 2023 Julian Hofer
 // Copyright (c) 2023 Huan Nguyen
 //
 // This program is free software: you can redistribute it and/or modify
@@ -18,24 +20,31 @@
 //
 // SPDX-License-Identifier: GPL-3.0-or-later
 
-use crate::about;
-use crate::deps::*;
-use crate::util::spawn;
-
 use adw::prelude::*;
 use adw::subclass::prelude::*;
-use glib::clone;
 
-use crate::config;
+use crate::deps::*;
 use crate::window::LpWindow;
+use crate::{about, config};
 
 mod imp {
     use super::*;
+    use crate::widgets::LpShyBin;
 
     // The basic struct that holds our
     // state and widgets
-    #[derive(Default, Debug)]
-    pub struct LpApplication {}
+    #[derive(Debug)]
+    pub struct LpApplication {
+        pub settings: gio::Settings,
+    }
+
+    impl Default for LpApplication {
+        fn default() -> Self {
+            Self {
+                settings: gio::Settings::new(config::APP_ID),
+            }
+        }
+    }
 
     // Sets up the basics for the GObject
     // The `#[glib::object_subclass] macro implements
@@ -65,7 +74,14 @@ mod imp {
 
     // Overrides GApplication vfuncs
     impl ApplicationImpl for LpApplication {
+        fn startup(&self) {
+            log::trace!("Startup");
+            self.parent_startup();
+            LpShyBin::ensure_type();
+        }
+
         fn activate(&self) {
+            log::trace!("Activate");
             let application = self.obj();
             let window = LpWindow::new(&*application);
             window.present();
@@ -73,6 +89,7 @@ mod imp {
 
         // Handles opening files from the command line or other applications
         fn open(&self, files: &[gio::File], _hint: &str) {
+            log::trace!("Open {} file(s)", files.len());
             let application = self.obj();
             let win = LpWindow::new(&*application);
             win.image_view().set_images_from_files(files.to_vec());
@@ -114,7 +131,7 @@ impl LpApplication {
             gio::ActionEntryBuilder::new("about")
                 .activate(|app: &Self, _, _| {
                     let app = app.clone();
-                    spawn(async move { app.show_about().await });
+                    glib::spawn_future_local(async move { app.show_about().await });
                 })
                 .build(),
             gio::ActionEntryBuilder::new("help")
@@ -141,14 +158,15 @@ impl LpApplication {
         self.set_accels_for_action("win.print", &["<Ctrl>P"]);
         self.set_accels_for_action("win.copy", &["<Ctrl>C"]);
         self.set_accels_for_action("win.trash", &["Delete", "KP_Delete"]);
+        self.set_accels_for_action("win.delete", &["<Shift>Delete", "<Shift>KP_Delete"]);
         self.set_accels_for_action("win.show-help-overlay", &["<Ctrl>question"]);
         self.set_accels_for_action("win.leave-fullscreen", &["Escape"]);
         self.set_accels_for_action("win.toggle-properties", &["F9", "<Alt>Return"]);
         self.set_accels_for_action("win.toggle-fullscreen", &["F11"]);
         self.set_accels_for_action("win.set-background", &["<Ctrl>F8"]);
 
-        self.set_accels_for_action("win.image-left", &["Left", "Page_Down"]);
-        self.set_accels_for_action("win.image-right", &["Right", "Page_Up"]);
+        self.set_accels_for_action("win.image-left-instant", &["Left", "Page_Down"]);
+        self.set_accels_for_action("win.image-right-instant", &["Right", "Page_Up"]);
         self.set_accels_for_action("win.first", &["Home"]);
         self.set_accels_for_action("win.last", &["End"]);
 
@@ -162,7 +180,7 @@ impl LpApplication {
         );
         self.set_accels_for_action("win.zoom-best-fit", &["0", "KP_0", "<Ctrl>0", "<Ctrl>KP_0"]);
         self.set_accels_for_action(
-            "win.zoom-in",
+            "win.zoom-in-cursor",
             &[
                 "<Ctrl>plus",
                 "plus",
@@ -173,7 +191,7 @@ impl LpApplication {
             ],
         );
         self.set_accels_for_action(
-            "win.zoom-out",
+            "win.zoom-out-cursor",
             &["<Ctrl>minus", "minus", "<Ctrl>KP_Subtract", "KP_Subtract"],
         );
 
@@ -199,14 +217,28 @@ impl LpApplication {
     }
 
     pub fn show_help(&self) {
-        spawn(clone!(@weak self as app => async move {
-            let context = app
-                .active_window()
-                .map(|w| gtk::prelude::WidgetExt::display(&w).app_launch_context());
-
-            if let Err(e) = gio::AppInfo::launch_default_for_uri_future("help:loupe", context.as_ref()).await {
+        let context = self
+            .active_window()
+            .map(|w| gtk::prelude::WidgetExt::display(&w).app_launch_context());
+        glib::spawn_future_local(async move {
+            if let Err(e) =
+                gio::AppInfo::launch_default_for_uri_future("help:loupe", context.as_ref()).await
+            {
                 log::error!("Failed to launch help: {}", e.message());
             }
-        }));
+        });
+    }
+
+    pub fn settings(&self) -> gio::Settings {
+        self.imp().settings.clone()
+    }
+}
+
+impl Default for LpApplication {
+    fn default() -> Self {
+        gio::Application::default()
+            .unwrap()
+            .downcast::<LpApplication>()
+            .unwrap()
     }
 }
