@@ -1,4 +1,4 @@
-// Copyright (c) 2023 Sophie Herold
+// Copyright (c) 2023-2024 Sophie Herold
 // Copyright (c) 2023 FineFindus
 //
 // This program is free software: you can redistribute it and/or modify
@@ -20,22 +20,21 @@
 //!
 //! It is not always feasible or desiareble to store the complete decoded
 //! image in the VRAM. [`TiledImage`] allows to compose the parts of
-//! the image currently viewed, out of smaller [`Tiles`].
+//! the image currently viewed, out of smaller [`Tile`]s.
 //!
 //! This is especially important for SVGs where every the image has to
 //! be re-genearted for each zoom level. It can also be used to allow
 //! showing large JPEGs etc where the complete decoded image would not
 //! even fit in the VRAM.
-use super::{DecoderUpdate, ImageDimensionDetails, UpdateSender};
-use crate::deps::*;
+use std::collections::{BTreeMap, VecDeque};
+use std::sync::Arc;
+use std::time::Duration;
 
 use arc_swap::ArcSwap;
 use gtk::prelude::*;
 
-use std::collections::BTreeMap;
-use std::collections::VecDeque;
-use std::sync::Arc;
-use std::time::Duration;
+use super::{DecoderUpdate, UpdateSender};
+use crate::deps::*;
 
 const ZOOM_SIGNIFICANT_DIGITS: i32 = 6;
 pub const MIN_TILE_SIZE: u16 = 500;
@@ -100,7 +99,7 @@ impl Tile {
         options: &RenderOptions,
     ) {
         // Multiply by scale factor to get the image scaled to `zoom` in physical pixels
-        let zoom = output_zoom as f32 * options.scale_factor as f32 / self.zoom();
+        let zoom = output_zoom as f32 * options.scaling as f32 / self.zoom();
         let area = round(&self.area().scale(zoom, zoom));
         let area_with_bleed = round(&self.area_with_bleed().scale(zoom, zoom));
 
@@ -109,7 +108,7 @@ impl Tile {
             return;
         }
 
-        // TODO: do not clip outer bounderies of the image
+        // TODO: do not clip outer boundaries of the image
         snapshot.push_clip(&area);
         // TODO: do not clip outer boundaries of the image
         if let Some(background_color) = &options.background_color {
@@ -215,7 +214,7 @@ pub struct TileLayer {
 #[derive(Debug, Clone)]
 pub struct RenderOptions {
     pub scaling_filter: gsk::ScalingFilter,
-    pub scale_factor: i32,
+    pub scaling: f64,
     pub background_color: Option<gdk::RGBA>,
 }
 
@@ -232,16 +231,13 @@ impl TiledImage {
         log::trace!("Rendering {} tiles", tiles.len());
 
         // Scale from application pixels back to physical pixels
-        snapshot.scale(
-            1. / options.scale_factor as f32,
-            1. / options.scale_factor as f32,
-        );
+        snapshot.scale(1. / options.scaling as f32, 1. / options.scaling as f32);
         // reverse to put least fitting tiles in the background
         for (_, tile) in tiles.iter().rev() {
             tile.add_to_snapshot(snapshot, zoom, options);
         }
         // reset scale
-        snapshot.scale(options.scale_factor as f32, options.scale_factor as f32);
+        snapshot.scale(options.scaling as f32, options.scaling as f32);
     }
 
     pub fn push(&mut self, tile: Tile) {
@@ -335,8 +331,8 @@ impl TiledImage {
 
     /// Viewport within zoom coordinates
     ///
-    /// preserve_area: Slightly larger than the viewport, area for which we want tiles
-    /// TODO: This is the most buggy function ever
+    /// preserve_area: Slightly larger than the viewport, area for which we want
+    /// tiles TODO: This is the most buggy function ever
     pub fn cleanup(&mut self, zoom: f64, preserve_area: graphene::Rect) {
         let zoom = zoom_normalize(zoom);
         let mut kept_tiles = Self::default();
@@ -466,7 +462,7 @@ impl Tiling {
 
     /// Returns relevant tiles
     ///
-    /// This gives what tiles are needed for convering the `preload_area`.
+    /// This gives what tiles are needed for covering the `preload_area`.
     /// Decoders use the [`TileInstructions`] to know what they have to render.
     pub fn relevant_tiles(&self, preload_area: &graphene::Rect) -> Vec<TileInstructions> {
         let original_width = self.scaled_dimensions().0 as f32;
@@ -576,76 +572,68 @@ impl RectExt for graphene::Rect {
     }
 }
 
-/// We always store [`FrameBuffer`]s in an [`ArcSwap`].
-/// This trait adds convenience functions for this.
-pub trait FrameBufferExt {
-    fn push(&self, tile: Tile);
-    fn push_tile(&self, tiling: Tiling, position: Coordinates, texture: gdk::Texture);
-    fn push_frame(&self, tile: Tile, dimensions: Coordinates, delay: Duration);
-    fn reset(&self);
-    fn get_layer_tiling_or_default(&self, zoom: f64, viewport: graphene::Rect) -> Tiling;
-    /// Return true if the next frame should be shown and removes the outdated frame
-    fn frame_timeout(&self, elapsed: Duration) -> bool;
-    /// Returns the number of currently buffered frames
-    fn n_frames(&self) -> usize;
-    fn set_original_dimensions(&self, size: Coordinates);
-    fn set_original_dimensions_full(
-        &self,
-        size: Coordinates,
-        dimension_details: ImageDimensionDetails,
-    );
-    fn set_update_sender(&self, sender: UpdateSender);
+#[derive(Debug, Default)]
+pub struct SharedFrameBuffer {
+    buffer: ArcSwap<FrameBuffer>,
 }
 
-impl FrameBufferExt for ArcSwap<FrameBuffer> {
-    fn push(&self, tile: Tile) {
-        self.rcu(|tiling_store| {
+impl std::ops::Deref for SharedFrameBuffer {
+    type Target = ArcSwap<FrameBuffer>;
+
+    fn deref(&self) -> &Self::Target {
+        &self.buffer
+    }
+}
+
+impl SharedFrameBuffer {
+    pub fn push(&self, tile: Tile) {
+        self.buffer.rcu(|tiling_store| {
             let mut new_store = (**tiling_store).clone();
             new_store.current().push(tile.clone());
             Arc::new(new_store)
         });
-        if let Some(updater) = &self.load().update_sender {
+        if let Some(updater) = &self.buffer.load().update_sender {
             updater.send(DecoderUpdate::Redraw);
         }
     }
 
-    fn push_tile(&self, tiling: Tiling, position: Coordinates, texture: gdk::Texture) {
-        self.rcu(|tiling_store| {
+    pub fn push_tile(&self, tiling: Tiling, position: Coordinates, texture: gdk::Texture) {
+        self.buffer.rcu(|tiling_store| {
             let mut new_store = (**tiling_store).clone();
             new_store
                 .current()
                 .push_tile(tiling, position, texture.clone());
             Arc::new(new_store)
         });
-        if let Some(updater) = &self.load().update_sender {
+        if let Some(updater) = &self.buffer.load().update_sender {
             updater.send(DecoderUpdate::Redraw);
         }
     }
 
-    fn push_frame(&self, tile: Tile, dimensions: Coordinates, delay: Duration) {
+    pub fn push_frame(&self, tile: Tile, dimensions: Coordinates, delay: Duration) {
         let mut store = TiledImage::default();
         store.push(tile);
         store.original_dimensions = Some(dimensions);
         store.delay = delay;
 
-        self.rcu(|tiling_store| {
+        self.buffer.rcu(|tiling_store| {
             let mut new_store = (**tiling_store).clone();
             new_store.images.push_back(store.clone());
             Arc::new(new_store)
         });
     }
 
-    fn reset(&self) {
-        self.rcu(|tiling_store| {
+    pub fn reset(&self) {
+        self.buffer.rcu(|tiling_store| {
             let mut new_store = (**tiling_store).clone();
             new_store.reset();
             Arc::new(new_store)
         });
     }
 
-    fn get_layer_tiling_or_default(&self, zoom: f64, viewport: graphene::Rect) -> Tiling {
+    pub fn get_layer_tiling_or_default(&self, zoom: f64, viewport: graphene::Rect) -> Tiling {
         let mut tiling = None;
-        self.rcu(|tiling_store| {
+        self.buffer.rcu(|tiling_store| {
             let mut new_store = (**tiling_store).clone();
             tiling = Some(
                 new_store
@@ -657,50 +645,50 @@ impl FrameBufferExt for ArcSwap<FrameBuffer> {
         tiling.unwrap()
     }
 
-    fn frame_timeout(&self, elapsed: Duration) -> bool {
+    /// Return true if the next frame should be shown and removes the outdated
+    /// frame
+    pub fn frame_timeout(&self, elapsed: Duration) -> bool {
         if self
+            .buffer
             .load()
             .images
             .front()
             .is_some_and(|next_frame| elapsed >= next_frame.delay)
         {
-            self.rcu(|tiling_store| {
-                let mut new_store = (**tiling_store).clone();
-                new_store.images.pop_front();
-                Arc::new(new_store)
-            });
+            self.next_frame();
             return true;
         }
 
         false
     }
 
-    fn n_frames(&self) -> usize {
-        self.load().images.len()
+    pub fn next_frame(&self) {
+        self.buffer.rcu(|tiling_store| {
+            let mut new_store = (**tiling_store).clone();
+            new_store.images.pop_front();
+            Arc::new(new_store)
+        });
     }
 
-    fn set_original_dimensions(&self, size: Coordinates) {
-        self.set_original_dimensions_full(size, Default::default());
+    /// Returns the number of currently buffered frames
+    pub fn n_frames(&self) -> usize {
+        self.buffer.load().images.len()
     }
 
-    fn set_original_dimensions_full(
-        &self,
-        size: Coordinates,
-        dimension_details: ImageDimensionDetails,
-    ) {
-        self.rcu(|tiling_store| {
+    pub fn set_original_dimensions(&self, size: Coordinates) {
+        self.buffer.rcu(|tiling_store| {
             let mut new_store = (**tiling_store).clone();
             new_store.current().original_dimensions = Some(size);
             Arc::new(new_store)
         });
 
-        if let Some(updater) = &self.load().update_sender {
-            updater.send(DecoderUpdate::Dimensions(dimension_details));
+        if let Some(updater) = &self.buffer.load().update_sender {
+            updater.send(DecoderUpdate::Dimensions);
         }
     }
 
-    fn set_update_sender(&self, sender: UpdateSender) {
-        self.rcu(|tiling_store| {
+    pub fn set_update_sender(&self, sender: UpdateSender) {
+        self.buffer.rcu(|tiling_store| {
             let mut new_store = (**tiling_store).clone();
             new_store.set_update_sender(sender.clone());
             Arc::new(new_store)
