@@ -1,4 +1,5 @@
-// Copyright (c) 2023 Sophie Herold
+// Copyright (c) 2023-2024 Sophie Herold
+// Copyright (c) 2023 Julian Hofer
 //
 // This program is free software: you can redistribute it and/or modify
 // it under the terms of the GNU General Public License as published by
@@ -17,19 +18,18 @@
 
 //! The print layout widget that also coordinates the print process
 
-use crate::decoder;
-use crate::deps::*;
-use crate::util::gettext::*;
-use crate::widgets::{LpImage, LpPrintPreview};
+use std::cell::{Cell, OnceCell, RefCell};
+use std::rc::Rc;
 
 use adw::prelude::*;
 use adw::subclass::prelude::*;
 use glib::Properties;
 use gtk::CompositeTemplate;
-use once_cell::sync::OnceCell;
 
-use std::cell::{Cell, RefCell};
-use std::rc::Rc;
+use crate::decoder::{self};
+use crate::deps::*;
+use crate::util::gettext::*;
+use crate::widgets::{LpImage, LpPrintPreview};
 
 #[derive(Default, Clone, Copy)]
 struct PageAlignment {
@@ -63,8 +63,8 @@ enum Status {
 
 /// Scope guard for non user ui changes
 ///
-/// Creates a context in which other signals know that changes are not user input.
-/// This avoids things like loops between value change signals.
+/// Creates a context in which other signals know that changes are not user
+/// input. This avoids things like loops between value change signals.
 #[derive(Default, Clone, Debug)]
 struct UiUpdates {
     disabled: Rc<Cell<bool>>,
@@ -92,7 +92,7 @@ impl From<&str> for PageAlignment {
             "left" => (HAlignment::Left, VAlignment::Middle),
             "right" => (HAlignment::Right, VAlignment::Middle),
             pos => {
-                log::error!("Unkown alignment '{pos}'");
+                log::error!("Unknown alignment '{pos}'");
                 (HAlignment::default(), VAlignment::default())
             }
         };
@@ -166,12 +166,12 @@ impl std::fmt::Display for Unit {
 }
 
 impl Unit {
-    fn factor(&self, dpi: f64, total_size: i32) -> f64 {
+    fn factor(&self, dpi: f64, total_size: f64) -> f64 {
         match self {
             Self::Centimeter => 2.54 / dpi,
             Self::Inch => 1. / dpi,
             Self::Pixel => 1.,
-            Self::Percent => 100. / total_size as f64,
+            Self::Percent => 100. / total_size,
         }
     }
 
@@ -211,7 +211,7 @@ mod imp {
     use super::*;
 
     #[derive(Debug, Default, CompositeTemplate, Properties)]
-    #[template(file = "../../data/gtk/print.ui")]
+    #[template(file = "print.ui")]
     #[properties(wrapper_type = super::LpPrint)]
     pub struct LpPrint {
         #[template_child]
@@ -248,6 +248,8 @@ mod imp {
         #[property(get, set, construct_only)]
         pub(super) image: OnceCell<LpImage>,
         #[property(get, set, construct_only)]
+        pub(super) preview_image: OnceCell<LpImage>,
+        #[property(get, set, construct_only)]
         pub(super) parent_window: OnceCell<gtk::Window>,
         #[property(get, set, construct_only)]
         pub(super) print_settings: OnceCell<gtk::PrintSettings>,
@@ -270,14 +272,13 @@ mod imp {
 
         fn class_init(klass: &mut Self::Class) {
             Self::bind_template(klass);
-            Self::Type::bind_template_callbacks(klass);
 
-            klass.install_action_async("print.print", None, |print, _, _| async move {
+            klass.install_action("print.print", None, |print, _, _| {
                 print.print();
             });
 
-            klass.install_action_async("print.back", None, |print, _, _| async move {
-                print.back().await;
+            klass.install_action("print.back", None, |print, _, _| {
+                print.back();
             });
 
             klass.install_property_action("print.orientation", "orientation");
@@ -368,7 +369,7 @@ mod imp {
                     imp.title.set_title(&gettext_f(
                         // Translators: {} is a placeholder for the filename
                         "Print “{}”",
-                        &[&basename],
+                        [basename],
                     ));
                     if let Some(printer) = operation.print_settings().and_then(|x| x.printer()) {
                         imp.title.set_subtitle(&printer);
@@ -392,7 +393,7 @@ mod imp {
                     let size_unit = obj.size_unit();
                     imp.width.set_value(
                         size_unit
-                            .round(obj.image().image_size().0 as f64 * obj.width_unit_factor()),
+                            .round(obj.original_size().0 as f64 * obj.width_unit_factor()),
                     );
 
                     // Default to inch for USA and Liberia
@@ -477,7 +478,6 @@ glib::wrapper! {
         @implements gtk::Buildable, gtk::ConstraintTarget;
 }
 
-#[gtk::template_callbacks]
 impl LpPrint {
     pub fn new(
         image: LpImage,
@@ -485,10 +485,20 @@ impl LpPrint {
         print_settings: Option<gtk::PrintSettings>,
         page_setup: Option<gtk::PageSetup>,
     ) -> Self {
+        let preview_image = LpImage::new_still();
+        preview_image.set_fixed_background_color(Some(gdk::RGBA::new(0.94, 0.94, 0.94, 1.)));
+        if let Some(file) = image.file() {
+            let preview_image_ = preview_image.clone();
+            glib::spawn_future_local(async move { preview_image_.load(&file).await });
+        } else {
+            log::error!("Trying to print image without file");
+        }
+
         let print_settings = print_settings.unwrap_or_default();
 
         let obj: Self = glib::Object::builder()
             .property("image", image)
+            .property("preview-image", preview_image)
             .property("print-settings", print_settings)
             .property("parent-window", parent_window)
             .build();
@@ -501,6 +511,20 @@ impl LpPrint {
         obj
     }
 
+    pub fn original_size(&self) -> (f64, f64) {
+        let image = self.image();
+        let metadata = image.metadata();
+        if let Some((w, h)) = metadata.dimensions_inch() {
+            let dpi = self.dpi();
+
+            ((w * dpi).round(), (h * dpi).round())
+        } else {
+            let (w, h) = self.image().image_size();
+
+            (w as f64, h as f64)
+        }
+    }
+
     /// Starts print preparation process by showing the print dialog
     pub fn run(&self) -> Result<(), glib::Error> {
         self.print_operation()
@@ -567,10 +591,10 @@ impl LpPrint {
             // Disable 'fill space' when manually changing size
             imp.fill_space.set_active(false);
 
-            let (orig_width, orig_height) = self.image().image_size();
+            let (orig_width, orig_height) = self.original_size();
             let width = self
                 .size_unit()
-                .round(imp.height.value() * orig_width as f64 / orig_height as f64);
+                .round(imp.height.value() * orig_width / orig_height);
             imp.width.set_value(width);
 
             self.draw_preview();
@@ -596,13 +620,20 @@ impl LpPrint {
 
     /// Returns user selected image height in pixels
     pub fn user_height(&self) -> f64 {
-        let (_, orig_height) = self.image().image_size();
+        let (_, orig_height) = self.original_size();
 
-        f64::max(1., orig_height as f64 * self.user_scale())
+        f64::max(1., orig_height * self.user_scale())
     }
 
     /// Returns scaling of the original image based on selected image size
     pub fn user_scale(&self) -> f64 {
+        let (orig_width, _) = self.original_size();
+
+        self.user_width() / orig_width
+    }
+
+    /// Returns scaling of the original image based on selected image size
+    pub fn user_scale_original(&self) -> f64 {
         let (orig_width, _) = self.image().image_size();
 
         self.user_width() / orig_width as f64
@@ -691,7 +722,7 @@ impl LpPrint {
 
         let size_unit = self.size_unit();
         let max_width = self.fill_space_width();
-        let max_height = self.paper_height() - 2. * self.user_margin_vertical();
+        let max_height = self.fill_space_height();
 
         imp.width.set_range(
             size_unit.ceil(self.width_unit_factor()),
@@ -708,17 +739,17 @@ impl LpPrint {
         let width1 = self.paper_width() - 2. * self.user_margin_horizontal();
 
         let height = self.paper_height() - 2. * self.user_margin_vertical();
-        let (orig_width, orig_height) = self.image().image_size();
-        let width2 = height * orig_width as f64 / orig_height as f64;
+        let (orig_width, orig_height) = self.original_size();
+        let width2 = height * orig_width / orig_height;
 
         f64::min(width1, width2)
     }
 
     /// Returns height that makes the image fill the page
     fn fill_space_height(&self) -> f64 {
-        let (orig_width, orig_height) = self.image().image_size();
+        let (orig_width, orig_height) = self.original_size();
 
-        self.fill_space_width() * orig_height as f64 / orig_width as f64
+        self.fill_space_width() * orig_height / orig_width
     }
 
     /// Updates width according to margins if fill space is activated
@@ -745,9 +776,19 @@ impl LpPrint {
 
     fn on_size_unit_changed(&self) {
         let imp = self.imp();
+
+        // Only show one size input for percentage
+        if matches!(self.size_unit(), Unit::Percent) {
+            imp.width.set_title(&gettext("_Scale"));
+            imp.height.set_visible(false);
+        } else {
+            imp.width.set_title(&gettext("_Width"));
+            imp.height.set_visible(true);
+        }
+
         let _ui_updates_disabled = self.disable_ui_updates();
 
-        let (orig_width, orig_height) = self.image().image_size();
+        let (orig_width, orig_height) = self.original_size();
 
         let unit = self.size_unit();
         let width_factor = unit.factor(self.dpi(), orig_width);
@@ -773,10 +814,11 @@ impl LpPrint {
 
     fn on_margin_unit_changed(&self) {
         let imp = self.imp();
+
         let _ui_updates_disabled = self.disable_ui_updates();
 
         let unit = self.margin_unit();
-        let unit_factor = unit.factor(self.dpi(), 1);
+        let unit_factor = unit.factor(self.dpi(), 1.);
 
         let update_factor = unit_factor / self.margin_unit_factor();
         let margin_horizontal = imp.margin_horizontal.value() * update_factor;
@@ -813,7 +855,7 @@ impl LpPrint {
     }
 
     /// Go back to print dialog from image layout
-    async fn back(&self) {
+    fn back(&self) {
         self.close();
         let print = Self::new(
             self.image(),
@@ -850,18 +892,18 @@ impl LpPrint {
 
         let cairo_dpi = print_context.dpi_x();
 
-        let (orig_width, orig_height) = image.image_size();
+        let (orig_width, orig_height) = self.original_size();
 
-        let texture_scale = self.user_width() / orig_width as f64;
+        let texture_scale = self.user_width() / orig_width;
         let cairo_scale = cairo_dpi / self.dpi();
 
-        let texture = if image.format().map_or(false, |x| x.is_svg()) {
+        let texture = if image.metadata().is_svg() {
             // Render SVG to exact needed sizes
             // TODO: This should be async
             decoder::formats::Svg::render_print(
                 &image.file().unwrap(),
-                (orig_width as f64 * texture_scale) as i32,
-                (orig_height as f64 * texture_scale) as i32,
+                (orig_width * texture_scale) as i32,
+                (orig_height * texture_scale) as i32,
             )
             .unwrap()
         } else {
