@@ -1,12 +1,13 @@
 // Copyright (c) 2020-2023 Christopher Davis
-// Copyright (c) 2022-2023 Sophie Herold
+// Copyright (c) 2022-2024 Sophie Herold
 // Copyright (c) 2022 Elton A Rodrigues
 // Copyright (c) 2022 Maximiliano Sandoval R
+// Copyright (c) 2023 Matteo Nardi
+// Copyright (c) 2023 Julian Hofer
 // Copyright (c) 2023 FineFindus
 // Copyright (c) 2023 qwel
 // Copyright (c) 2023 Huan Thieu Nguyen
 // Copyright (c) 2023 Sabri Ünal
-// Copyright (c) 2023 Lubosz Sarnecki
 //
 // This program is free software: you can redistribute it and/or modify
 // it under the terms of the GNU General Public License as published by
@@ -23,20 +24,21 @@
 //
 // SPDX-License-Identifier: GPL-3.0-or-later
 
-use crate::deps::*;
-use crate::util::gettext::*;
-
-use crate::util::spawn;
-use adw::prelude::*;
-use adw::subclass::prelude::*;
-use glib::clone;
-use gtk::CompositeTemplate;
+mod controls;
 
 use std::cell::{Cell, OnceCell, RefCell};
 use std::path::{Path, PathBuf};
 
+use adw::prelude::*;
+use adw::subclass::prelude::*;
+use glib::clone;
+use gtk::{CompositeTemplate, Widget};
+
+use crate::application::LpApplication;
 use crate::config;
-use crate::util::{self, Direction, Position};
+use crate::deps::*;
+use crate::util::gettext::*;
+use crate::util::{Direction, Position};
 use crate::widgets::{LpDragOverlay, LpImage, LpImageView, LpPropertiesView};
 
 /// Show window after X milliseconds even if image dimensions are not known yet
@@ -50,6 +52,8 @@ const HIDE_CONTROLS_ANIMATION_DURATION: u32 = 1000;
 const HIDE_CONTROLS_IDLE_TIMEOUT: u64 = 3000;
 
 mod imp {
+    use gio::glib::VariantTy;
+
     use super::*;
 
     // To use composite templates, you need
@@ -67,7 +71,7 @@ mod imp {
     // we'd need to have a `new()` function in the
     // `impl ObjectSubclass for $TYPE` section.
     #[derive(Default, Debug, CompositeTemplate)]
-    #[template(file = "../data/gtk/window.ui")]
+    #[template(file = "window.ui")]
     pub struct LpWindow {
         // Template children are used with the
         // TemplateChild<T> wrapper, where T is the
@@ -81,6 +85,8 @@ mod imp {
         #[template_child]
         pub(super) properties_button: TemplateChild<gtk::ToggleButton>,
         #[template_child]
+        pub(super) primary_menu: TemplateChild<gtk::MenuButton>,
+        #[template_child]
         pub(super) toast_overlay: TemplateChild<adw::ToastOverlay>,
         #[template_child]
         pub(super) stack: TemplateChild<gtk::Stack>,
@@ -120,7 +126,16 @@ mod imp {
             // bind_template() is a function generated by the
             // CompositeTemplate macro to bind all children at once.
             Self::bind_template(klass);
-            Self::Type::bind_template_callbacks(klass);
+
+            klass.add_binding(gdk::Key::c, gdk::ModifierType::CONTROL_MASK, |window| {
+                if window.has_metadata_selected() {
+                    // Pass on to normal copy handler to copy selected metadata
+                    glib::Propagation::Proceed
+                } else {
+                    window.copy_image();
+                    glib::Propagation::Stop
+                }
+            });
 
             // Set up actions
             klass.install_action("win.toggle-fullscreen", None, move |win, _, _| {
@@ -128,26 +143,42 @@ mod imp {
             });
 
             klass.install_action("win.next", None, move |win, _, _| {
-                win.imp().image_view.navigate(Direction::Forward);
+                win.imp().image_view.navigate(Direction::Forward, true);
             });
 
             klass.install_action("win.previous", None, move |win, _, _| {
-                win.imp().image_view.navigate(Direction::Back);
+                win.imp().image_view.navigate(Direction::Back, true);
             });
 
             klass.install_action("win.image-right", None, move |win, _, _| {
                 if win.direction() == gtk::TextDirection::Rtl {
-                    win.imp().image_view.navigate(Direction::Back);
+                    win.imp().image_view.navigate(Direction::Back, true);
                 } else {
-                    win.imp().image_view.navigate(Direction::Forward);
+                    win.imp().image_view.navigate(Direction::Forward, true);
+                }
+            });
+
+            klass.install_action("win.image-right-instant", None, move |win, _, _| {
+                if win.direction() == gtk::TextDirection::Rtl {
+                    win.imp().image_view.navigate(Direction::Back, false);
+                } else {
+                    win.imp().image_view.navigate(Direction::Forward, false);
                 }
             });
 
             klass.install_action("win.image-left", None, move |win, _, _| {
                 if win.direction() == gtk::TextDirection::Rtl {
-                    win.imp().image_view.navigate(Direction::Forward);
+                    win.imp().image_view.navigate(Direction::Forward, true);
                 } else {
-                    win.imp().image_view.navigate(Direction::Back);
+                    win.imp().image_view.navigate(Direction::Back, true);
+                }
+            });
+
+            klass.install_action("win.image-left-instant", None, move |win, _, _| {
+                if win.direction() == gtk::TextDirection::Rtl {
+                    win.imp().image_view.navigate(Direction::Forward, false);
+                } else {
+                    win.imp().image_view.navigate(Direction::Back, false);
                 }
             });
 
@@ -159,18 +190,30 @@ mod imp {
                 win.imp().image_view.jump(Position::Last);
             });
 
-            klass.install_action("win.zoom-out", None, move |win, _, _| {
-                win.zoom_out();
+            klass.install_action("win.zoom-out-cursor", None, move |win, _, _| {
+                win.zoom_out_cursor();
             });
 
-            klass.install_action("win.zoom-in", None, move |win, _, _| {
-                win.zoom_in();
+            klass.install_action("win.zoom-out-center", None, move |win, _, _| {
+                win.zoom_out_center();
             });
 
-            klass.install_action("win.zoom-to-exact", Some("d"), move |win, _, level| {
-                win.zoom_to_exact(level.unwrap().get().unwrap());
+            klass.install_action("win.zoom-in-cursor", None, move |win, _, _| {
+                win.zoom_in_cursor();
             });
 
+            klass.install_action("win.zoom-in-center", None, move |win, _, _| {
+                win.zoom_in_center();
+            });
+
+            klass.install_action(
+                "win.zoom-to-exact",
+                Some(VariantTy::DOUBLE),
+                move |win, _, level| {
+                    win.zoom_to_exact(level.unwrap().get().unwrap());
+                },
+            );
+
             klass.install_action("win.zoom-best-fit", None, move |win, _, _| {
                 win.zoom_best_fit();
             });
@@ -225,19 +268,27 @@ mod imp {
                 win.print();
             });
 
-            klass.install_action("win.copy", None, move |win, _, _| {
-                win.copy();
+            klass.install_action("win.copy-image", None, move |win, _, _| {
+                win.copy_image();
             });
 
             klass.install_action_async("win.trash", None, |win, _, _| async move {
                 win.trash().await;
             });
 
-            klass.install_action("win.show-toast", Some("(si)"), move |win, _, var| {
-                if let Some((ref toast, i)) = var.and_then(|v| v.get::<(String, i32)>()) {
-                    win.show_toast(toast, adw::ToastPriority::__Unknown(i));
-                }
+            klass.install_action_async("win.delete", None, |win, _, _| async move {
+                win.delete().await;
             });
+
+            klass.install_action(
+                "win.show-toast",
+                Some(VariantTy::TUPLE),
+                move |win, _, var| {
+                    if let Some((ref toast, i)) = var.and_then(|v| v.get::<(String, i32)>()) {
+                        win.show_toast(toast, adw::ToastPriority::__Unknown(i));
+                    }
+                },
+            );
         }
 
         fn instance_init(obj: &glib::subclass::InitializingObject<Self>) {
@@ -247,17 +298,21 @@ mod imp {
 
     impl ObjectImpl for LpWindow {
         fn constructed(&self) {
-            let obj = self.obj();
-
             self.parent_constructed();
 
+            let obj = self.obj();
+
             self.forward_click_gesture
                 .connect_pressed(clone!(@weak obj => move |_,_,_,_| {
-                    obj.image_view().navigate(Direction::Forward);
+                    obj.image_view().navigate(Direction::Forward, true);
                 }));
             self.backward_click_gesture
                 .connect_pressed(clone!(@weak obj => move |_,_,_,_| {
-                    obj.image_view().navigate(Direction::Back);
+                    obj.image_view().navigate(Direction::Back, true);
+                }));
+            self.properties_button
+                .connect_toggled(clone!(@weak obj => move |_| {
+                    obj.on_properties_button_toggled();
                 }));
 
             if config::PROFILE == ".Devel" {
@@ -273,7 +328,24 @@ mod imp {
                 glib::clone!(@weak obj => move || if !obj.is_visible() { obj.present() }),
             );
 
-            obj.on_images_available();
+            obj.on_current_page_changed();
+            obj.on_fullscreen_changed();
+
+            self.image_view.connect_current_page_notify(
+                glib::clone!(@weak obj => move |_| obj.on_current_page_changed()),
+            );
+
+            obj.connect_fullscreened_notify(
+                glib::clone!(@weak obj => move |_| obj.on_fullscreen_changed()),
+            );
+
+            obj.connect_map(|win| {
+                win.resize_default();
+            });
+
+            self.properties_button.connect_active_notify(
+                glib::clone!(@weak obj => move |_| obj.update_headerbar_style()),
+            );
 
             let gesture_click = gtk::GestureClick::new();
             gesture_click
@@ -299,39 +371,47 @@ mod imp {
                 obj.on_zoom_status_changed()
             }));
 
-            let win = &*obj;
+            current_image_signals.connect_local(
+                "metadata-changed",
+                true,
+                glib::clone!(@weak obj => @default-return None, move |_| {
+                    obj.update_title();
+                    None
+                }),
+            );
+
             current_image_signals.connect_closure(
                 "notify::best-fit",
                 false,
                 // `closure_local!` is similar to `clone`, but you use `@watch` instead of clone.
                 // `@watch` means that this signal will be disconnected when the watched object
                 // is dropped.
-                glib::closure_local!(@watch win => move |_: &LpImage, _: &glib::ParamSpec| {
-                    win.on_zoom_status_changed();
+                glib::closure_local!(@watch obj => move |_: &LpImage, _: &glib::ParamSpec| {
+                    obj.on_zoom_status_changed();
                 }),
             );
 
             current_image_signals.connect_closure(
                 "notify::is-max-zoom",
                 false,
-                glib::closure_local!(@watch win => move |_: &LpImage, _: &glib::ParamSpec| {
-                    win.on_zoom_status_changed();
+                glib::closure_local!(@watch obj => move |_: &LpImage, _: &glib::ParamSpec| {
+                    obj.on_zoom_status_changed();
                 }),
             );
 
             current_image_signals.connect_closure(
-                "notify::image-size",
+                "notify::image-size-available",
                 false,
-                glib::closure_local!(@watch win => move |_: &LpImage, _: &glib::ParamSpec| {
-                    win.image_size_ready();
+                glib::closure_local!(@watch obj => move |_: &LpImage, _: &glib::ParamSpec| {
+                    obj.image_size_available();
                 }),
             );
 
             current_image_signals.connect_closure(
                 "notify::error",
                 false,
-                glib::closure_local!(@watch win => move |_: &LpImage, _: &glib::ParamSpec| {
-                    win.image_error();
+                glib::closure_local!(@watch obj => move |_: &LpImage, _: &glib::ParamSpec| {
+                    obj.image_error();
                 }),
             );
 
@@ -339,17 +419,9 @@ mod imp {
                 glib::clone!(@weak obj => move |_| obj.on_top_bar_style_notify()),
             );
 
-            self.image_view.connect_notify_local(
-                Some("current-page"),
-                glib::clone!(@weak obj => move |_, _| {
-                    obj.on_images_available();
-                }),
-            );
-
             // action win.previous status
-            self.image_view.connect_notify_local(
-                Some("is-previous-available"),
-                glib::clone!(@weak obj => move |_, _| {
+            self.image_view.connect_is_previous_available_notify(
+                glib::clone!(@weak obj => move |_| {
                     obj.action_set_enabled(
                         "win.previous",
                         obj.imp().image_view.is_previous_available(),
@@ -358,20 +430,18 @@ mod imp {
             );
 
             // action win.next status
-            self.image_view.connect_notify_local(
-                Some("is-next-available"),
-                glib::clone!(@weak obj => move |_, _| {
+            self.image_view
+                .connect_is_next_available_notify(glib::clone!(@weak obj => move |_| {
                     obj.action_set_enabled(
                         "win.next",
                         obj.imp().image_view.is_next_available(),
                     );
-                }),
-            );
+                }));
 
             // Make widgets visible when the focus moves
             obj.connect_move_focus(|obj, _| {
                 obj.show_controls();
-                obj.queue_hide_controls();
+                obj.schedule_hide_controls();
             });
 
             self.status_page
@@ -440,7 +510,6 @@ glib::wrapper! {
         @implements gio::ActionMap, gio::ActionGroup, gtk::Native;
 }
 
-#[gtk::template_callbacks]
 impl LpWindow {
     pub fn new<A: IsA<gtk::Application>>(app: &A) -> Self {
         glib::Object::builder().property("application", app).build()
@@ -450,12 +519,28 @@ impl LpWindow {
         self.set_fullscreened(fullscreen);
     }
 
-    fn zoom_out(&self) {
-        self.imp().image_view.zoom_out();
+    fn zoom_out_cursor(&self) {
+        if let Some(image) = self.imp().image_view.current_image() {
+            image.zoom_out_cursor();
+        }
     }
 
-    fn zoom_in(&self) {
-        self.imp().image_view.zoom_in();
+    fn zoom_out_center(&self) {
+        if let Some(image) = self.imp().image_view.current_image() {
+            image.zoom_out_center();
+        }
+    }
+
+    fn zoom_in_cursor(&self) {
+        if let Some(image) = self.imp().image_view.current_image() {
+            image.zoom_in_cursor();
+        }
+    }
+
+    fn zoom_in_center(&self) {
+        if let Some(image) = self.imp().image_view.current_image() {
+            image.zoom_in_center();
+        }
     }
 
     fn zoom_to_exact(&self, level: f64) {
@@ -481,12 +566,9 @@ impl LpWindow {
 
         let filter_supported_formats = gtk::FileFilter::new();
         filter_supported_formats.set_name(Some(&gettext("Supported image formats")));
-        for mime_type in glycin::image_formats().await {
+        for mime_type in glycin::supported_mime_types().await {
             filter_supported_formats.add_mime_type(&mime_type);
         }
-        // TODO: Update when SVG moves to glycin as well
-        filter_supported_formats.add_mime_type("image/svg+xml");
-        filter_supported_formats.add_mime_type("image/svg+xml-compressed");
 
         let filter_all_files = gtk::FileFilter::new();
         filter_all_files.set_name(Some(&gettext("All files")));
@@ -502,6 +584,14 @@ impl LpWindow {
             .modal(true)
             .build();
 
+        chooser.set_initial_folder(
+            self.image_view()
+                .current_image()
+                .and_then(|x| x.file())
+                .and_then(|x| x.parent())
+                .as_ref(),
+        );
+
         if let Ok(selected) = chooser.open_multiple_future(Some(self)).await {
             let images: Vec<_> = selected
                 .into_iter()
@@ -553,7 +643,21 @@ impl LpWindow {
         }
     }
 
-    fn copy(&self) {
+    /// Returns true if some text in metadata is currently selected
+    fn has_metadata_selected(&self) -> bool {
+        if let Some(focus_widget) = self.focus_widget() {
+            if focus_widget.is_ancestor(&*self.imp().properties_view) {
+                if let Ok(label) = focus_widget.downcast::<gtk::Label>() {
+                    return label.selection_bounds().is_some();
+                }
+            }
+        }
+
+        false
+    }
+
+    /// Copy image to clipboard or metadata text if selected
+    fn copy_image(&self) {
         let imp = self.imp();
 
         if let Err(e) = imp.image_view.copy() {
@@ -587,17 +691,16 @@ impl LpWindow {
                     .build();
                 toast.connect_button_clicked(glib::clone!(@weak self as win => move |_| {
                     let path = path.clone();
-                    spawn(async move {
+                    glib::spawn_future_local(async move {
+                        win.image_view()
+                            .set_trash_restore(Some(gio::File::for_path(&path)));
                         let result = crate::util::untrash(&path).await;
-                        match result {
-                            Ok(()) => win.image_view().set_images_from_files(vec![gio::File::for_path(&path)]),
-                            Err(err) => {
-                                log::error!("Failed to untrash {path:?}: {err}");
-                                win.show_toast(
-                                    gettext("Failed to restore image from trash"),
-                                    adw::ToastPriority::High,
-                                );
-                            }
+                        if let Err(err) = result {
+                            log::error!("Failed to untrash {path:?}: {err}");
+                            win.show_toast(
+                                gettext("Failed to restore image from trash"),
+                                adw::ToastPriority::High,
+                            );
                         }
                     });
                 }));
@@ -605,7 +708,7 @@ impl LpWindow {
             }
             Err(err) => {
                 if Some(gio::IOErrorEnum::NotSupported) == err.kind::<gio::IOErrorEnum>() {
-                    self.delete(&path).await;
+                    self.delete_future(&path).await;
                 } else {
                     log::error!("Failed to delete file {path:?}: {err}");
                     self.show_toast(
@@ -617,17 +720,28 @@ impl LpWindow {
         }
     }
 
+    /// Delete action called
+    async fn delete(&self) {
+        let image_view = self.image_view();
+        let Some(path) = image_view.current_file().and_then(|x| x.path()) else {
+            log::error!("No file to delete");
+            return;
+        };
+
+        self.delete_future(&path).await;
+    }
+
     /// Permanently delete image
     ///
-    /// Fallback for when trash not available
-    async fn delete(&self, path: &Path) {
+    /// Fallback for when trash not available or explicit call with shortcut
+    async fn delete_future(&self, path: &Path) {
         let dialog = adw::MessageDialog::builder()
             .modal(true)
             .transient_for(self)
             .heading(gettext("Permanently Delete Image?"))
             .body(gettext_f(
-                "The image “{}” can only be deleted permanently.",
-                &[&PathBuf::from(&path.file_name().unwrap_or_default())
+                "After deleting the image “{}” it will be permanently lost.",
+                [PathBuf::from(&path.file_name().unwrap_or_default())
                     .display()
                     .to_string()],
             ))
@@ -671,26 +785,44 @@ impl LpWindow {
     }
 
     pub fn set_actions_enabled(&self, enabled: bool) {
-        self.action_set_enabled("win.open-with", enabled);
-        self.action_set_enabled("win.set-background", enabled);
-        self.action_set_enabled("win.toggle-fullscreen", enabled);
-        self.action_set_enabled("win.print", enabled);
-        self.action_set_enabled("win.rotate_cw", enabled);
-        self.action_set_enabled("win.rotate_ccw", enabled);
-        self.action_set_enabled("win.copy", enabled);
-        self.action_set_enabled("win.zoom-best-fit", enabled);
-        self.action_set_enabled("win.zoom-to-exact", enabled);
-        self.action_set_enabled("win.toggle-properties", enabled);
+        const ACTIONS: &[&str] = &[
+            "win.open-with",
+            "win.set-background",
+            "win.toggle-fullscreen",
+            "win.print",
+            "win.rotate_cw",
+            "win.rotate_ccw",
+            "win.copy-image",
+            "win.zoom-best-fit",
+            "win.zoom-to-exact",
+            "win.toggle-properties",
+        ];
+
+        for action in ACTIONS {
+            self.action_set_enabled(action, enabled);
+        }
     }
 
-    /// Handles change in availability of images
-    fn on_images_available(&self) {
+    /// Handles change in image and availability of images
+    fn on_current_page_changed(&self) {
         let imp = self.imp();
+        let was_showing_image =
+            imp.stack.visible_child().as_ref() == Some(imp.image_view.upcast_ref::<Widget>());
+        let current_page = imp.image_view.current_page();
 
-        let shows_image = imp.image_view.current_page().is_some();
+        // HeaderBar style
+        self.update_headerbar_style();
 
-        imp.properties_button.set_sensitive(shows_image);
-        self.set_actions_enabled(shows_image);
+        // Window title
+        self.update_title();
+
+        // Properties view
+        let current_image = current_page.as_ref().map(|x| x.image());
+        imp.properties_view.set_image(current_image.as_ref());
+
+        let has_image = current_page.is_some();
+
+        self.set_actions_enabled(has_image);
         self.action_set_enabled(
             "win.trash",
             imp.image_view
@@ -698,19 +830,59 @@ impl LpWindow {
                 .is_some_and(|file| file.path().is_some()),
         );
 
-        if shows_image {
-            imp.stack.set_visible_child(&*imp.image_view);
-            imp.image_view.grab_focus();
-            self.queue_hide_controls();
+        if has_image {
+            // Properties buttons was not sensitive before
+            if !imp.properties_button.is_sensitive() {
+                let settings = LpApplication::default().settings();
+                // Pickup config for setting it's state
+                imp.properties_button
+                    .set_active(settings.boolean("show-properties"));
+            }
+            if !was_showing_image {
+                imp.stack.set_visible_child(&*imp.image_view);
+                imp.image_view.grab_focus();
+                self.show_controls();
+                self.schedule_hide_controls();
+            }
         } else {
+            imp.properties_button.set_active(false);
             imp.stack.set_visible_child(&*imp.status_page);
             imp.status_page.grab_focus();
             // Leave fullscreen since status page has no controls to leave it
             self.set_fullscreened(false);
         }
+
+        imp.properties_button.set_sensitive(has_image);
     }
 
-    pub fn image_size_ready(&self) {
+    /// When the image-properties sidebar is displayed or hidden, we should
+    /// update the "show-properties" setting.
+    fn on_properties_button_toggled(&self) {
+        let imp = self.imp();
+        // When no image is shown, we skip this update as the sidebar should always be
+        // hidden. This can happen when deleting a picture.
+        if imp.image_view.current_page().is_none() {
+            return;
+        }
+        let settings = LpApplication::default().settings();
+        let result = settings.set_boolean("show-properties", imp.properties_button.is_active());
+        if let Err(err) = result {
+            log::warn!("Failed to save show-properties state, {}", err);
+        }
+    }
+
+    pub fn update_title(&self) {
+        let title = self
+            .imp()
+            .image_view
+            .current_image()
+            .and_then(|x| x.metadata().file_name())
+            .unwrap_or_else(|| gettext("Image Viewer"));
+
+        self.set_title(Some(&title));
+    }
+
+    pub fn image_size_available(&self) {
         // if visible for whatever reason, don't do any resize
         if self.is_visible() {
             return;
@@ -722,11 +894,20 @@ impl LpWindow {
             .current_page()
             .map(|page| page.image());
 
-        if image.is_some_and(|img| img.image_size() > (0, 0)) {
+        if image.is_some_and(|img| img.image_size_available()) {
             log::debug!("Showing window because image size is ready");
+            self.present();
+        }
+    }
+
+    pub fn resize_default(&self) {
+        if self
+            .image_view()
+            .current_image()
+            .is_some_and(|img| img.image_size_available())
+        {
             // this let's the window determine the default size from LpImage's natural size
             self.set_default_size(-1, -1);
-            self.present();
         }
     }
 
@@ -755,192 +936,24 @@ impl LpWindow {
             .map(|image| !image.is_max_zoom())
             .unwrap_or_default();
 
-        self.action_set_enabled("win.zoom-out", can_zoom_out);
-        self.action_set_enabled("win.zoom-in", can_zoom_in);
+        self.action_set_enabled("win.zoom-out-cursor", can_zoom_out);
+        self.action_set_enabled("win.zoom-out-center", can_zoom_out);
+        self.action_set_enabled("win.zoom-in-cursor", can_zoom_in);
+        self.action_set_enabled("win.zoom-in-center", can_zoom_in);
     }
 
-    fn set_control_opacity(&self, opacity: f64, hiding: bool) {
-        self.image_view().controls_box_start().set_opacity(opacity);
-        self.image_view().controls_box_end().set_opacity(opacity);
+    fn on_fullscreen_changed(&self) {
+        self.imp()
+            .image_view
+            .on_fullscreen_changed(self.is_fullscreened());
 
-        if self.is_headerbar_flat() {
-            self.headerbar().set_opacity(opacity);
-        } else {
-            self.headerbar().set_opacity(1.);
-        }
+        self.update_headerbar_style();
 
-        if self.is_fullscreened() && hiding && opacity < 0.9 {
-            self.set_cursor(gdk::Cursor::from_name("none", None).as_ref());
-        } else {
-            self.set_cursor(None);
-        }
-    }
-
-    fn controls_opacity(&self) -> f64 {
-        self.image_view().controls_box_start().opacity()
-    }
-
-    /// Animation to show controls
-    fn show_controls_animation(&self) -> &adw::TimedAnimation {
-        self.imp().show_controls_animation.get_or_init(|| {
-            let target = adw::CallbackAnimationTarget::new(glib::clone!(
-                @weak self as obj => move |opacity| obj.set_control_opacity(opacity, false)
-            ));
-
-            adw::TimedAnimation::builder()
-                .duration(SHOW_CONTROLS_ANIMATION_DURATION)
-                .widget(&self.image_view())
-                .target(&target)
-                .value_to(1.)
-                .build()
-        })
-    }
-
-    /// Animation to hide controls
-    fn hide_controls_animation(&self) -> &adw::TimedAnimation {
-        self.imp().hide_controls_animation.get_or_init(|| {
-            let target = adw::CallbackAnimationTarget::new(glib::clone!(
-                @weak self as obj => move |opacity| obj.set_control_opacity(opacity, true)
-            ));
-
-            adw::TimedAnimation::builder()
-                .duration(HIDE_CONTROLS_ANIMATION_DURATION)
-                .widget(&self.image_view())
-                .target(&target)
-                .value_to(0.)
-                .build()
-        })
-    }
-
-    /// Queue a fade animation to play after `HIDE_CONTROLS_IDLE_TIMEOUT` of inactivity
-    fn queue_hide_controls(&self) {
-        self.dequeue_hide_controls();
-
-        let new_timeout = glib::timeout_add_local_once(
-            std::time::Duration::from_millis(HIDE_CONTROLS_IDLE_TIMEOUT),
-            glib::clone!(@weak self as win => move || {
-                win.imp().hide_controls_timeout.take();
-                win.hide_controls();
-            }),
-        );
-
-        self.imp().hide_controls_timeout.replace(Some(new_timeout));
-    }
-
-    fn dequeue_hide_controls(&self) {
-        if let Some(current_timeout) = self.imp().hide_controls_timeout.take() {
-            current_timeout.remove();
-        }
-    }
-
-    fn is_controls_visible(&self) -> bool {
-        if self.hide_controls_animation().state() == adw::AnimationState::Playing {
-            return false;
-        }
-
-        self.image_view().controls_box_start().opacity() == 1.
-            || self.show_controls_animation().state() == adw::AnimationState::Playing
-    }
-
-    /// Start animation to show controls
-    fn show_controls(&self) {
-        if !self.is_controls_visible() {
-            self.hide_controls_animation().pause();
-            self.show_controls_animation()
-                .set_value_from(self.controls_opacity());
-            self.show_controls_animation().play();
-        }
-    }
-
-    /// Start animation to hide controls
-    fn hide_controls(&self) {
-        if self.is_controls_visible() {
-            self.show_controls_animation().pause();
-            self.hide_controls_animation()
-                .set_value_from(self.controls_opacity());
-            self.hide_controls_animation().play();
-        }
-    }
-
-    /// Whether or not the window is showing images
-    fn is_showing_image(&self) -> bool {
-        let imp = self.imp();
-        imp.stack.visible_child().as_ref() == Some(imp.image_view.upcast_ref())
-    }
-
-    // In the LpWindow UI file we define a `gtk::Expression`s
-    // that is a closure. This closure takes the current `gio::File`
-    // and processes it to return a window title.
-    //
-    // In this function we chain `Option`s with `and_then()` in order
-    // to handle optional results with a fallback, without needing to
-    // have multiple `match` or `if let` branches, and without needing
-    // to unwrap.
-    #[template_callback]
-    fn window_title(&self, file: Option<&gio::File>) -> String {
-        // ensure that templates are initialized
-        if file.is_none() {
-            gettext("Image Viewer")
-        } else {
-            self.imp()
-                .image_view
-                .current_file()
-                .and_then(|f| util::get_file_display_name(&f)) // If the file exists, get display name
-                .unwrap_or_else(|| gettext("Image Viewer")) // Return that or the default if there's nothing
-        }
-    }
-
-    fn on_click(&self) {
-        self.show_controls();
-
-        if self.can_hide_controls() {
-            self.queue_hide_controls();
-        } else {
-            self.dequeue_hide_controls();
-        }
-    }
-
-    fn on_motion(&self, pointer_position: (f64, f64)) {
-        let imp = self.imp();
-
-        // Check if position really changed since swipe gesture sends change event with same position.
-        // Also, don't connect this to "leave" because swipe also sends fake "leave" events when
-        // the widget under the cursor changes.
-        if imp.pointer_position.get() == pointer_position {
-            return;
-        }
-        imp.pointer_position.set(pointer_position);
-
-        self.show_controls();
-
-        if self.can_hide_controls() {
-            self.queue_hide_controls();
-        } else {
-            self.dequeue_hide_controls();
-        }
-    }
-
-    /// Only hide controls if cursor not over controls and there is an image shown
-    fn can_hide_controls(&self) -> bool {
-        !self
-            .image_view()
-            .controls_box_start_events()
-            .contains_pointer()
-            && !self
-                .image_view()
-                .controls_box_end_events()
-                .contains_pointer()
-            && self.is_showing_image()
-            && (!self.is_headerbar_flat() || !self.imp().headerbar_events.contains_pointer())
-    }
-
-    #[template_callback]
-    fn on_fullscreened(&self) {
         if !self.is_fullscreened() {
             self.set_cursor(None);
             self.show_controls();
         }
-        self.queue_hide_controls();
+        self.schedule_hide_controls();
     }
 
     fn is_headerbar_flat(&self) -> bool {
@@ -959,25 +972,23 @@ impl LpWindow {
         }
     }
 
-    #[template_callback]
-    fn extend_to_top(&self, fullscreened: bool, show_properties: bool, show_menu: bool) -> bool {
-        fullscreened && !show_properties && !show_menu
+    fn is_content_extended_to_top(&self) -> bool {
+        self.is_fullscreened() && !self.imp().properties_button.is_active()
     }
 
-    #[template_callback]
-    fn top_bar_style(
-        &self,
-        fullscreened: bool,
-        show_properties: bool,
-        show_menu: bool,
-        current_image: Option<LpImage>,
-    ) -> adw::ToolbarStyle {
+    fn update_headerbar_style(&self) {
+        let imp = self.imp();
+
+        let extend = self.is_content_extended_to_top();
+        imp.toolbar_view.set_extend_content_to_top_edge(extend);
+
         // Flat headerbar for empty state and fullscreen
-        if current_image.is_none() || self.extend_to_top(fullscreened, show_properties, show_menu) {
+        let toolbar_style = if imp.image_view.current_image().is_none() || extend {
             adw::ToolbarStyle::Flat
         } else {
             // Use the border variant of raised to avoid shadows over images
-            adw::ToolbarStyle::__Unknown(2)
-        }
+            adw::ToolbarStyle::RaisedBorder
+        };
+        imp.toolbar_view.set_top_bar_style(toolbar_style);
     }
 }
